/* -*-C-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of an OpenGL plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

*/

/* Adapters for the GL graphics library. */

#include <GL/glu.h>
#include <gdk/gdkx.h>
#include <malloc.h>
#include <mit-scheme.h>
#include "gl-shim.h"

void
gl_clear_color (double *color)
{
  glClearColor (color[0], color[1], color[2], color[3]);
}

void
gl_light (GLenum light, GLenum pname, double *params)
{
  int i,len = flovec_length (params);
  GLfloat *fvec = malloc (len * sizeof (GLfloat));
  if (!fvec)
    error_external_return ();
  for (i = 0; i < len; i++) {
    fvec[i] = params[i];
  }
  glLightfv (light, pname, fvec);
  free (fvec);
}

void
gl_light_model_v (GLenum pname, double *params)
{
  int i, len = flovec_length (params);
  GLfloat *fvec = malloc (len * sizeof (GLfloat));
  if (!fvec)
    error_external_return ();
  for (i = 0; i < len; i++) {
    fvec[i] = params[i];
  }
  glLightModelfv (pname, fvec);
  free (fvec);
}

void
gl_material (GLenum face, GLenum pname, double *params)
{
  int i, len = flovec_length (params);
  GLfloat *fvec = malloc (len * sizeof (GLfloat));
  if (!fvec)
    error_external_return ();
  for (i = 0; i < len; i++) {
    fvec[i] = params[i];
  }
  glMaterialfv (face, pname, fvec);
  free (fvec);
}

void
glu_look_at (double *eye, double *center, double *up)
{
  gluLookAt (eye[0], eye[1], eye[2],
	     center[0], center[1], center[2],
	     up[0], up[1], up[2]);
}

static int gdk_visual_get_visual_class (GdkVisual *v);

GdkVisual*
glx_find_gdkvisual (GdkWindow *window, XVisualInfo *visinfo)
{
  GList *list = gdk_screen_list_visuals (gdk_window_get_screen (window));
  GdkVisual *found = NULL;
  int match_class = visinfo->class;
  int match_depth = visinfo->depth;
  GList *scan = list;
  while (scan) {
    GdkVisual *v = scan->data;
    if (gdk_visual_get_visual_class (v) == match_class
	&& gdk_visual_get_depth (v) == match_depth)
      {
	found = v;
	break;
      }
    scan = scan->next;
  }
  g_list_free (list);
  if (!found)
    error_external_return ();
  return (found);
}

static int
gdk_visual_get_visual_class (GdkVisual *v)
{
  GdkVisualType type = gdk_visual_get_visual_type (v);
  switch (type)
    {
    case GDK_VISUAL_STATIC_GRAY:
      return (StaticGray);
    case GDK_VISUAL_GRAYSCALE:
      return (GrayScale);
    case GDK_VISUAL_STATIC_COLOR:
      return (StaticColor);
    case GDK_VISUAL_PSEUDO_COLOR:
      return (PseudoColor);
    case GDK_VISUAL_TRUE_COLOR:
      return (TrueColor);
    case GDK_VISUAL_DIRECT_COLOR:
      return (DirectColor);
    default:
      return (-1);
    }
}

int
gdk_window_xid (GdkWindow *window)
{
  return (GDK_WINDOW_XID (window));
}

Display *
gdk_window_xdisplay (GdkWindow *window)
{
  return (GDK_WINDOW_XDISPLAY (window));
}

int
gdk_window_screen_num (GdkWindow *window)
{
  return (gdk_screen_get_number (gdk_window_get_screen (window)));
}
