#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of an OpenGL plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; The GLX Gears Demo

;;(declare (no-type-checks) (no-range-checks))

(declare (reduce-operator (sin flo:sin)
			  (cos flo:cos)
			  (sqrt flo:sqrt)
			  (= flo:=)
 			  (< flo:<)
 			  (> flo:>)
			  (+ flo:+)
			  (- flo:- (null-value 0. single) (group left))
			  (* flo:*)
			  (/ flo:/ (null-value 1. single) (group left))))

(define (make-glxgears-demo)
  (make-glxgears-demo-device 400 400 "gl-glxgears.scm"))

(define (make-glxgears-demo-device width height title)
  (with-glib-lock
   (lambda ()
     (let ((window (gtk-window-new 'toplevel)))
       (gtk-window-set-title window title)
       (set-gtk-window-delete-event-callback!
	window (lambda (w) (%trace "closed "w) 0))
       (gtk-container-set-border-width window 5)
       (let ((widget (%make-glxgears-demo width height)))
	 (gtk-container-add window widget)
	 (gtk-widget-show-all window)
	 widget)))))

(define-class (<glxgears-demo>
	       (constructor %make-glxgears-demo () (width height)))
    (<glx-widget>)
  (view-rotx define standard initial-value 20.)
  (view-roty define standard initial-value 30.)
  (angle define standard initial-value 0.)
  (gears define standard initial-value #f)

  (shape define standard initial-value '(0 . 0))
  ;; This "visibility" slot was intended to reflect when my screen is
  ;; blanked, because the animation loop spins up to more than 300fps.
  ;; (Please help stop the spinning!)  Why so few visibility-notify
  ;; events from Gtk lately? ?
  (visibility define standard initial-value #f)
  (animate? define standard initial-value #t)
  (animation-thread define standard)
  (animation-halt define standard)
  ;; For smooth rotation at any frame rate?
  (frame-start define standard initial-value #f)

  ;; For frame rate reports:
  (report-fps? define standard initial-value #t)
  (frame-count define standard initial-value 0)
  (frame-count-start define standard initial-value #f))

(define-method initialize-instance ((widget <glxgears-demo>) width height)
  (call-next-method widget width height)
  (set-gtk-widget-draw-callback! widget draw-callback)
  (set-fix-widget-visibility-notify-handler! widget visibility-notify-handler)
  (set-fix-widget-key-press-handler! widget key-press-handler)
  (set-fix-widget-motion-handler! widget motion-handler)
  (set-fix-widget-button-handler! widget 'press button-handler)
  (set-fix-widget-button-handler! widget 'release button-handler)
  (set-fix-widget-button-handler! widget 'double-press button-handler)
  (set-fix-widget-button-handler! widget 'triple-press button-handler)

  (set-fix-widget-enter-notify-handler! widget enter-notify-handler)
  (set-fix-widget-leave-notify-handler! widget leave-notify-handler)
  (set-fix-widget-focus-change-handler! widget focus-change-handler)

  (make-animation-thread widget))

(define-method fix-widget-new-geometry-callback ((widget <glxgears-demo>))
  (call-next-method widget)
  (%trace "fix-widget-new-geometry-callback <glxgears-demo> "widget))

(define-method fix-widget-realize-callback ((widget <glxgears-demo>))
  (call-next-method widget)
  (%trace "fix-widget-realize-callback <glxgears-demo> "widget))

(define-method fix-widget-unrealize-callback ((widget <glxgears-demo>))
  (call-next-method widget)
  (%trace "fix-widget-unrealize-callback <glxgears-demo> "widget
	  " "(gtk-widget-destroyed? widget))
  ;; Is this necessary when the context is about to be (already?) destroyed?
  #;(let ((gears (glxgears-demo-gears widget)))
    (with-glx-widget widget
      (lambda ()
	(gl:delete-lists (car gears) 1)
	(gl:delete-lists (cadr gears) 1)
	(gl:delete-lists (caddr gears) 1))))
  (halt-animation-thread widget))

(define (draw-callback widget cr)
  (declare (ignore cr))
  (%trace2 "draw-callback "widget)
  (with-glx-widget widget
    (lambda ()
      (if (not (glxgears-demo-gears widget))
	  (init widget))
      (draw widget)
      (%trace2 "  swap-buffers")
      (glx:swap-buffers widget)))
  (wake-animation-thread widget)
  (%trace2 "draw-callback "widget" done"))

(define (enter-notify-handler widget)
  (%trace "enter-notify-handler "widget)
  #f)

(define (leave-notify-handler widget)
  (%trace "leave-notify-handler "widget)
  #f)

(define (focus-change-handler widget in?)
  (%trace "focus-change-handler "widget" "in?)
  #f)

(define (visibility-notify-handler widget how)
  (%trace "visibility-notify-handler "widget" "how)
  (let ((old (glxgears-demo-visibility widget)))
    (set-glxgears-demo-visibility! widget how)
    (if (and (eq? old 'obscured) (not (eq? how 'obscured)))
	(wake-animation-thread widget)))
  #t)

(define (key-press-handler widget key bits)
  (%trace "key-press-handler "widget" "key" "bits)
  ;;(declare (ignore bits))
  (case key
    ((#\escape #\q #\Q)
     (gtk-widget-destroy (gtk-widget-parent widget)))
    ((#\? #\h #\H #\c-h |F1|)
     (popup-help widget)
     #t)
    ((#\a #\A)
     (if (glxgears-demo-animate? widget)
	 (begin
	   (%trace "toggle animation off")
	   (set-glxgears-demo-animate?! widget #f))
	 (begin
	   (%trace "toggle animation on")
	   (set-glxgears-demo-frame-count-start! widget #f)
	   (set-glxgears-demo-animate?! widget #t)
	   (wake-animation-thread widget))))
    ((#\i #\I)
     (if (glxgears-demo-report-fps? widget)
	 (begin
	   (%trace "fps reporting off")
	   (set-glxgears-demo-report-fps?! widget #f))
	 (begin
	   (%trace "fps reporting on")
	   (set-glxgears-demo-frame-count-start! widget #f)
	   (set-glxgears-demo-report-fps?! widget #t))))	   
    ((|Up|) (rotx! widget 5.) (gtk-widget-queue-draw widget))
    ((|Down|) (rotx! widget -5.) (gtk-widget-queue-draw widget))
    ((|Left|) (roty! widget 5.) (gtk-widget-queue-draw widget))
    ((|Right|) (roty! widget -5.) (gtk-widget-queue-draw widget)))
  #t)

(define-integrable (rotx! widget incr)
  (set-glxgears-demo-view-rotx!
   widget (+ incr (glxgears-demo-view-rotx widget))))

(define-integrable (roty! widget incr)
  (set-glxgears-demo-view-roty!
   widget (+ incr (glxgears-demo-view-roty widget))))

(define help-window #f)

(define (popup-help widget)
  (declare (ignore widget))
  (if (or (not help-window)
	  (gtk-widget-destroyed? help-window))
      (let ((window (gtk-window-new 'toplevel))
	    (label (gtk-label-new
		    "Key - GLX Gears demo command.
ESC - Quit.
Up/Down/Left/Right - Orbit.
A - Toggle animation.
I - Toggle frame rate reports.")))
	(gtk-window-set-type-hint window 'utility)
	(gtk-window-set-accept-focus window #f)
	(gtk-window-set-title window "Help")
	(gtk-container-set-border-width window 20)
	(gtk-container-add window label)
	(set! help-window window)
	(gtk-widget-show-all window)))
  (gtk-window-present help-window))

(define (motion-handler widget modifiers x y)
  (%trace2 "motion-handler "widget" "modifiers" "x"x"y)
  #f)

(define (button-handler widget name button modifiers x y)
  (%trace "button-handler "widget" "name" "button" "modifiers" "x"x"y)
  #f)

(define (make-animation-thread widget)
  (let ((thread
	 (create-thread
	     #f
	   (lambda ()
	     (call-with-current-continuation
	      (lambda (halt)
		(set-glxgears-demo-animation-halt! widget halt)
		(%trace "animation started")
		(let loop ()

		  ;; Sleep when not realized?, not animate?, or obscured.
		  (with-thread-events-blocked
		   (lambda ()
		     (let ((visibility (glxgears-demo-visibility widget)))
		       (if (or (not (glx-widget-realized? widget))
			       (not (glxgears-demo-animate? widget))
			       (eq? 'obscured visibility))
			   (begin
			     (if (and %trace? (eq? 'obscured visibility))
				 (%trace2
				  "animation sleeping while obscured...")
				 (%trace2 "animation sleeping..."))
			     (suspend-current-thread)
			     (%trace2 "animation awake"))
			   (%trace2 "animation rolling")))))

		  (if (and (glx-widget-realized? widget)
			   (glxgears-demo-animate? widget)
			   (not (eq? 'obscured
				     (glxgears-demo-visibility widget))))
		      (draw-frame widget)
		      (%trace2 "animation skip!"))
		  (%trace2 "animation loop")
		  (loop))))
	     (%trace "animation done")))))
    (set-glxgears-demo-animation-thread! widget thread)
    thread))

(define (wake-animation-thread widget)
  (signal-thread-event
   (glxgears-demo-animation-thread widget)
   (lambda ()
     (%trace2 "wake animation")
     unspecific)))

(define (halt-animation-thread widget)
  (signal-thread-event
   (glxgears-demo-animation-thread widget)
   (lambda ()
     ((glxgears-demo-animation-halt widget) unspecific))))

(define (draw-frame widget)
  (%trace2 "draw-frame")

  (if (glxgears-demo-animate? widget)
      (let ((now (real-time-clock))
	    (start (glxgears-demo-frame-start widget))
	    (count (glxgears-demo-frame-count widget))
	    (angle (glxgears-demo-angle widget)))
	(set-glxgears-demo-frame-start! widget now)    
	(set-glxgears-demo-angle! widget
				  (let ((dt (if (not start)
						0.
						(internal-time/ticks->seconds
						 (int:- now start)))))
				    (+ angle (* 70. dt))))
	(set-glxgears-demo-frame-count! widget (int:1+ count))))

  (with-thread-events-blocked
   (lambda ()
     (%trace2 "draw-frame: request draw")
     (with-glib-lock (lambda () (gtk-widget-queue-draw widget)))
     (%trace2 "draw-frame: sleep")
     (suspend-current-thread)
     (%trace2 "draw-frame: awake")))

  (if (and (glxgears-demo-animate? widget)
	   (glxgears-demo-report-fps? widget))
      (let ((now (glxgears-demo-frame-start widget))
	    (count (glxgears-demo-frame-count widget))
	    (start (glxgears-demo-frame-count-start widget)))
	(if start
	    (let ((seconds
		   (->flonum (internal-time/ticks->seconds (fix:- now start)))))
	      (if (>= seconds 5.0)
		  (let ((fps (/ (->flonum count) seconds)))
		    (define-integrable (%.1f n)
		      (number->string (/ (round (* 10. n)) 10.)))
		    (define-integrable (%.3f n)
		      (number->string (/ (round (* 1000. n)) 1000.)))
		    (fresh-line)
		    (for-each display (list "; glxgears: "count" frames"
					    " in "(%.1f seconds)" seconds"
					    " = "(%.3f fps)" FPS\n"))
		    (set-glxgears-demo-frame-count-start! widget now)
		    (set-glxgears-demo-frame-count! widget 0))))
	    (begin
	      (set-glxgears-demo-frame-count-start! widget now)
	      (set-glxgears-demo-frame-count! widget 0))))))

(define (draw widget)
  (%trace2 "  draw "widget)
  (let ((angle (glxgears-demo-angle widget))
	(gears (glxgears-demo-gears widget))
	(view-rotx (glxgears-demo-view-rotx widget))
	(view-roty (glxgears-demo-view-roty widget)))
    (let ((gear1 (car gears))
	  (gear2 (cadr gears))
	  (gear3 (caddr gears)))
      (draw-gears angle gear1 gear2 gear3 view-rotx view-roty))))

(define (draw-gears angle gear1 gear2 gear3 view-rotx view-roty)
  (%trace2 "    draw-gears "
	   angle" "gear1" "gear2" "gear3" "view-rotx" "view-roty)
  (gl:clear 'color-buffer 'depth-buffer)
  (gl:push-matrix)
  (gl:rotate view-rotx 1. 0. 0.)
  (gl:rotate view-roty 0. 1. 0.)
  #;(gl:rotate view-rotz 0. 0. 1.)

  (gl:push-matrix)
  (gl:translate -3. -2. 0.)
  (gl:rotate angle 0. 0. 1.)
  (gl:call-list gear1)
  (gl:pop-matrix)

  (gl:push-matrix)
  (gl:translate 3.1 -2. 0.)
  (gl:rotate (- (* -2.0 angle) 9.) 0. 0. 1.)
  (gl:call-list gear2)
  (gl:pop-matrix)

  (gl:push-matrix)
  (gl:translate -3.1 4.2 0.)
  (gl:rotate (- (* -2.0 angle) 25.) 0. 0. 1.)
  (gl:call-list gear3)
  (gl:pop-matrix)

  (gl:pop-matrix))

(define (reshape widget)
  (let ((geo (fix-widget-geometry widget))
	(w.h (glxgears-demo-shape widget)))
    (let ((width (fix-rect-width geo))
	  (height (fix-rect-height geo)))
      (if (not (and (fix:= (car w.h) width)
		    (fix:= (cdr w.h) height)))
	  (begin
	    (%trace "  reshape "width" "height" "widget)
		(gl:viewport 0 0 width height)
		(let ((h (/ (->flonum height) (->flonum width))))
		  (gl:matrix-mode 'projection)
		  (gl:load-identity)
		  (gl:frustum -1. 1. (- h) h 5. 60.))

		(gl:matrix-mode 'modelview)
		(gl:load-identity)
		(gl:translate 0. 0. -40.)
	    (set-glxgears-demo-frame-start! widget #f)
	    (set-glxgears-demo-frame-count! widget 0)
	    (set-glxgears-demo-shape! widget (cons width height)))
	  (%trace "  reshape "width" "height" "widget" no change")))))

(define (init widget)
  (%trace "init "widget)
  (reshape widget)
      (let ((pos (flo:4d 5.0 5.0 10.0 0.0))
	    (red (color 0.8 0.1 0.0 1.0))
	    (green (color 0.0 0.8 0.2 1.0))
	    (blue (color 0.2 0.2 1.0 1.0)))
	(%trace "light")
	(gl:light 'light0 'position pos)
	(gl:enable 'cull-face)
	(gl:enable 'lighting)
	(gl:enable 'light0)
	(gl:enable 'depth-test)

	;; make the gears
	(let ((gear1 (gl:gen-lists 1)))
	  (%trace "gear1 = "gear1)
	  (gl:new-list gear1 'compile)
	  (gl:material 'front 'ambient-and-diffuse red)
	  (draw-gear 1.0 4.0 1.0 20. 0.7)
	  (gl:end-list)

	  (let ((gear2 (gl:gen-lists 1)))
	    (%trace "gear2 = "gear2)
	    (gl:new-list gear2 'compile)
	    (gl:material 'front 'ambient-and-diffuse green)
	    (draw-gear 0.5 2.0 2.0 10. 0.7)
	    (gl:end-list)

	    (let ((gear3 (gl:gen-lists 1)))
	      (%trace "gear3 = "gear3)
	      (gl:new-list gear3 'compile)
	      (gl:material 'front 'ambient-and-diffuse blue)
	      (draw-gear 1.3 2.0 0.5 10. 0.7)
	      (gl:end-list)

	      (gl:enable 'normalize)
	      (set-glxgears-demo-gears! widget (list gear1 gear2 gear3)))))))

(define (draw-gear inner-radius		; radius of hole at center
		   outer-radius		; radius at center of teeth
		   width		; width of gear
		   teeth		; number of teeth
		   tooth-depth)		; depth of tooth
  (%trace "draw-gear "inner-radius" "outer-radius" "width" "teeth" "tooth-depth)
  (let ((r0 inner-radius)
	(r1 (- outer-radius (/ tooth-depth 2.)))
	(r2 (+ outer-radius (/ tooth-depth 2.)))
	(2pi/teeth (/ 2pi teeth))
	(width/2 (* width .5))
	(-width/2 (* (- width) .5)))
    (let ((da (/ 2pi/teeth 4.)))
      (let ((2da (* 2. da))
	    (3da (* 3. da)))

	(gl:shade-model 'flat)
	(gl:normal (flo:3d 0. 0. 1.))

	;; draw front face
	(gl:begin 'quad-strip)
	(do ((i 0. (+ i 1.)))
	    ((> i teeth))
	  (let ((angle (* i 2pi/teeth)))
	    (gl:vertex3 (* r0 (cos angle)) (* r0 (sin angle)) width/2)
	    (gl:vertex3 (* r1 (cos angle)) (* r1 (sin angle)) width/2)
	    (if (< i teeth)
		(let ((angl4 (+ angle 3da)))
		  (gl:vertex3 (* r0 (cos angle)) (* r0 (sin angle)) width/2)
		  (gl:vertex3 (* r1 (cos angl4)) (* r1 (sin angl4)) width/2)))))
	(gl:end)

	;; draw front sides of teeth
	(gl:begin 'quads)
	(do ((i 0. (+ i 1.)))
	    ((= i teeth))
	  (let ((angle (* i 2pi/teeth)))
	    (let ((angl1 angle))
	      (gl:vertex3 (* r1 (cos angl1)) (* r1 (sin angl1)) width/2))
	    (let ((angl2 (+ angle da)))
	      (gl:vertex3 (* r2 (cos angl2)) (* r2 (sin angl2)) width/2))
	    (let ((angl3 (+ angle 2da)))
	      (gl:vertex3 (* r2 (cos angl3)) (* r2 (sin angl3)) width/2))
	    (let ((angl4 (+ angle 3da)))
	      (gl:vertex3 (* r1 (cos angl4)) (* r1 (sin angl4)) width/2))))
	(gl:end)

	(gl:normal (flo:3d 0. 0. -1.))

	;; draw back face
	(gl:begin 'quad-strip)
	(do ((i 0. (+ i 1.)))
	    ((> i teeth))
	  (let ((angle (* i 2pi/teeth)))
	    (gl:vertex3 (* r1 (cos angle)) (* r1 (sin angle)) -width/2)
	    (gl:vertex3 (* r0 (cos angle)) (* r0 (sin angle)) -width/2)
	    (if (< i teeth)
		(let ((angl4 (+ angle 3da)))
		  (gl:vertex3 (* r1 (cos angl4)) (* r1 (sin angl4)) -width/2)
		  (gl:vertex3 (* r0 (cos angle)) (* r0 (sin angle)) -width/2)
		  ))))
	(gl:end)

	;; draw back sides of teeth
	(gl:begin 'quads)
	(do ((i 0. (+ i 1.)))
	    ((= i teeth))
	  (let ((angle (* i 2pi/teeth)))
	    (let ((angl4 (+ angle 3da)))
	      (gl:vertex3 (* r1 (cos angl4)) (* r1 (sin angl4)) -width/2))
	    (let ((angl3 (+ angle 2da)))
	      (gl:vertex3 (* r2 (cos angl3)) (* r2 (sin angl3)) -width/2))
	    (let ((angl2 (+ angle da)))
	      (gl:vertex3 (* r2 (cos angl2)) (* r2 (sin angl2)) -width/2))
	    (let ((angl1 angle))
	      (gl:vertex3 (* r1 (cos angl1)) (* r1 (sin angl1)) -width/2))))
	(gl:end)

	;; draw outward faces of teeth
	(gl:begin 'quad-strip)
	(do ((i 0. (+ i 1.)))
	    ((= i teeth))
	  (let ((angle (* i 2pi/teeth)))
	    (let ((angl2 (+ angle da))
		  (angl3 (+ angle 2da))
		  (angl4 (+ angle 3da)))
	      (gl:vertex3 (* r1 (cos angle)) (* r1 (sin angle)) width/2)
	      (gl:vertex3 (* r1 (cos angle)) (* r1 (sin angle)) -width/2)
	      (let ((u (- (* r2 (cos angl2)) (* r1 (cos angle))))
		    (v (- (* r2 (sin angl2)) (* r1 (sin angle)))))
		(let ((len (sqrt (+ (* u u) (* v v)))))
		  (gl:normal (flo:3d (/ v len) (- (/ u len)) 0.))))
	      (gl:vertex3 (* r2 (cos angl2)) (* r2 (sin angl2)) width/2)
	      (gl:vertex3 (* r2 (cos angl2)) (* r2 (sin angl2)) -width/2)
	      (gl:normal (flo:3d (cos angle) (sin angle) 0.))
	      (gl:vertex3 (* r2 (cos angl3)) (* r2 (sin angl3)) width/2)
	      (gl:vertex3 (* r2 (cos angl3)) (* r2 (sin angl3)) -width/2)
	      (let ((u (- (* r1 (cos angl4)) (* r2 (cos angl3))))
		    (v (- (* r1 (sin angl4)) (* r2 (sin angl3)))))
		(gl:normal (flo:3d v (- u) 0.)))
	      (gl:vertex3 (* r1 (cos angl4)) (* r1 (sin angl4)) width/2)
	      (gl:vertex3 (* r1 (cos angl4)) (* r1 (sin angl4)) -width/2)
	      (gl:normal (flo:3d (cos angle) (sin angle) 0.)))))
	(gl:vertex3 (* r1 (cos 0.)) (* r1 (sin 0.)) width/2)
	(gl:vertex3 (* r1 (cos 0.)) (* r1 (sin 0.)) -width/2)
	(gl:end)

	(gl:shade-model 'smooth)

	;; draw inside radius cylinder
	(gl:begin 'quad-strip)
	(do ((i 0. (+ i 1.)))
	    ((> i teeth))
	  (let ((angle (* i 2pi/teeth)))
	    (gl:normal (flo:3d (- (cos angle)) (- (sin angle)) 0.))
	    (gl:vertex3 (* r0 (cos angle)) (* r0 (sin angle)) -width/2)
	    (gl:vertex3 (* r0 (cos angle)) (* r0 (sin angle)) width/2)))
	(gl:end)))))

(define-integrable (gl:vertex3 x y z)
  (gl:vertex (flo:3d x y z)))

(define-integrable 2pi (* 8. (flo:atan2 1. 1.)))

(define-integrable (flo:3d x y z)
  (let ((v (flo:vector-cons 3)))
    (flo:vector-set! v 0 x)
    (flo:vector-set! v 1 y)
    (flo:vector-set! v 2 z)
    v))

(define-integrable (flo:4d r g b a)
  (let ((v (flo:vector-cons 4)))
    (flo:vector-set! v 0 r)
    (flo:vector-set! v 1 g)
    (flo:vector-set! v 2 b)
    (flo:vector-set! v 3 a)
    v))

(define-integrable color flo:4d)

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ args ...)
     (if %trace? (outf-error "; glxgears: " args ... "\n")))))

(define %trace2? #f)

(define-syntax %trace2
  (syntax-rules ()
    ((_ args ...)
     (if %trace2? (outf-error "; glxgears: " args ... "\n")))))