#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of an OpenGL plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; A few gl wraps.
;;; package: (gl)

(C-include "gl")

(define (gl:shade-model model)
  (guarantee-current 'gl:shade-model)
  (C-call "glShadeModel"
	  (case model
	    ((flat) (C-enum "GL_FLAT"))
	    ((smooth) (C-enum "GL_SMOOTH"))
	    (else (error "gl:shade-model: Unknown model:" model)))))

(define (gl:clear-color color)
  (guarantee-current 'gl:clear-color)
  (guarantee-color color 'gl:clear-color)
  (C-call "gl_clear_color" color))

(define (gl:clear-depth depth)
  (guarantee-current 'gl:clear-depth)
  (guarantee-gl-depth depth 'gl:clear-depth)
  (C-call "glClearDepth" depth))

(define (->capability cap operator)
  (case cap
    ((depth-test) (C-enum "GL_DEPTH_TEST"))
    ((cull-face) (C-enum "GL_CULL_FACE"))
    ((light0) (c-enum "GL_LIGHT0"))
    ((light1) (C-enum "GL_LIGHT1"))
    ((light2) (C-enum "GL_LIGHT2"))
    ((light3) (C-enum "GL_LIGHT3"))
    ((light4) (C-enum "GL_LIGHT4"))
    ((light5) (C-enum "GL_LIGHT5"))
    ((light6) (C-enum "GL_LIGHT6"))
    ((light7) (C-enum "GL_LIGHT7"))
    ((lighting) (C-enum "GL_LIGHTING"))
    ((normalize) (C-enum "GL_NORMALIZE"))
    ((rescale-normal) (C-enum "GL_RESCALE_NORMAL"))
    ((color-material) (C-enum "GL_COLOR_MATERIAL"))
    (else (error:wrong-type-argument cap "a GL capability" operator))))

(define (gl:enable capability)
  (guarantee-current 'gl:enable)
  (C-call "glEnable" (->capability capability 'gl:enable)))

(define (gl:disable capability)
  (guarantee-current 'gl:disable)
  (C-call "glDisable" (->capability capability 'gl:disable)))

(define (gl:depth-func function)
  (guarantee-current 'gl:depth-func)
  (C-call "glDepthFunc"
	  (case function
	    ((lequal) (C-enum "GL_LEQUAL"))
	    (else (error "Unknown glDepthFunc function:" function)))))

(define (gl:blend-func sfactor dfactor)
  (guarantee-current 'gl:blend-func)
  (let ((s (->blend-factor sfactor 'gl:blend-func))
	(d (->blend-factor dfactor 'gl:blend-func)))
    (C-call "glBlendFunc" s d)))

(define (->blend-factor f op)
  (case f
    ((zero) (C-enum "GL_ZERO"))
    ((one) (C-enum "GL_ONE"))
    ((src-color) (C-enum "GL_SRC_COLOR"))
    ((one-minus-src-color) (C-enum "GL_ONE_MINUS_SRC_COLOR"))
    ((dst-color) (C-enum "GL_DST_COLOR"))
    ((one-minus-dst-color) (C-enum "GL_ONE_MINUS_DST_COLOR"))
    ((src-alpha) (C-enum "GL_SRC_ALPHA"))
    ((one-minus-src-alpha) (C-enum "GL_ONE_MINUS_SRC_ALPHA"))
    ((dst-alpha) (C-enum "GL_DST_ALPHA"))
    ((one-minus-dst-alpha) (C-enum "GL_ONE_MINUS_DST_ALPHA"))
    ((constant-color) (C-enum "GL_CONSTANT_COLOR"))
    ((one-minus-constant-color) (C-enum "GL_ONE_MINUS_CONSTANT_COLOR"))
    ((constant-alpha) (C-enum "GL_CONSTANT_ALPHA"))
    ((one-minus-constant-alpha) (C-enum "GL_ONE_MINUS_CONSTANT_ALPHA"))
    ((src-alpha-saturate) (C-enum "GL_SRC_ALPHA_SATURATE"))
    (else (error:wrong-type-argument f "GL blend factor" op))))

(define (gl:cull-face mode)
  (guarantee-current 'gl:cull-face)
  (C-call "glCullFace"
	  (case mode
	    ((front) (C-enum "GL_FRONT"))
	    ((back) (C-enum "GL_BACK"))
	    ((front-and-back) (C-enum "GL_FRONT_AND_BACK"))
	    (else (error "Unknown glCullFace mode:" mode)))))

(define (gl:hint target mode)
  (guarantee-current 'gl:hint)
  (C-call "glHint"
	  (case target
	    ((perspective-correction) (C-enum "GL_PERSPECTIVE_CORRECTION_HINT"))
	    (else (error "Unknown glHint target:" target)))
	  (case mode
	    ((nicest) (C-enum "GL_NICEST"))
	    (else (error "Unknown glHint mode:" mode)))))

(define (gl:color-material face mode)
  (C-call "glColorMaterial"
	  (case face
	    ((front) (C-enum "GL_FRONT"))
	    ((front-and-back) (C-enum "GL_FRONT_AND_BACK"))
	    ((back) (C-enum "GL_BACK"))
	    (else (error "Unknown glColorMaterial face:" face)))
	  (case mode
	    ((emission) (C-enum "GL_EMISSION"))
	    ((ambient) (C-enum "GL_AMBIENT"))
	    ((diffuse) (C-enum "GL_DIFFUSE"))
	    ((specular) (C-enum "GL_SPECULAR"))
	    ((ambient-and-diffuse) (C-enum "GL_AMBIENT_AND_DIFFUSE"))
	    (else (error "Unknown glColorMaterial mode:" mode)))))

(define (gl:clear . bits)
  (guarantee-current 'gl:clear)
  (C-call "glClear"
	  (reduce + 0 (map (lambda (bit)
			     (case bit
			       ((color-buffer)
				(C-enum "GL_COLOR_BUFFER_BIT"))
			       ((depth-buffer)
				(C-enum "GL_DEPTH_BUFFER_BIT"))
			       ((stencil-buffer)
				(C-enum "GL_STENCIL_BUFFER_BIT"))
			       (else (error "Unknwon glClear bit:" bit))))
			   bits))))

(define (gl:load-identity)
  (guarantee-current 'gl:load-identity)
  (C-call "glLoadIdentity"))

(define (gl:scale kx ky kz)
  (guarantee-current 'gl:scale)
  (guarantee flo:flonum? kx 'gl:scale)
  (guarantee flo:flonum? ky 'gl:scale)
  (guarantee flo:flonum? kz 'gl:scale)
  (C-call "glScaled" kx ky kz))

(define (gl:begin mode)
  (guarantee-current 'gl:begin)
  (C-call "glBegin"
	  (case mode
	    ((points) (C-enum "GL_POINTS"))
	    ((lines) (C-enum "GL_LINES"))
	    ((line-loop) (C-enum "GL_LINE_LOOP"))
	    ((line-strip) (C-enum "GL_LINE_STRIP"))
	    ((triangles) (C-enum "GL_TRIANGLES"))
	    ((triangle-strip) (C-enum "GL_TRIANGLE_STRIP"))
	    ((triangle-fan) (C-enum "GL_TRIANGLE_FAN"))
	    ((quads) (C-enum "GL_QUADS"))
	    ((quad-strip) (C-enum "GL_QUAD_STRIP"))
	    ((polygon) (C-enum "GL_POLYGON"))
	    (else (error "Unknown glBegin mode:" mode)))))

(define (gl:color color)
  (guarantee-current 'gl:color)
  (guarantee-color color 'gl:color)
  (C-call "glColor4dv" color))

(define (gl:vertex point)
  (guarantee-current 'gl:vertex)
  (guarantee-3d point 'gl:vertex)
  (C-call "glVertex3dv" point))

(define (gl:end)
  (guarantee-current 'gl:end)
  (C-call "glEnd"))

(define (gl:call-list lst)
  (guarantee-current 'gl:call-list)
  (guarantee integer? lst 'gl:call-list)
  (C-call "glCallList" lst))

(define (gl:draw-buffer buffer)
  (guarantee-current 'gl:draw-buffer)
  (C-call "glDrawBuffer"
	  (case buffer
	    ((back-left) (C-enum "GL_BACK_LEFT"))
	    ((back-right) (C-enum "GL_BACK_RIGHT"))
	    (else (error "gl:draw-buffer: Unknown buffer:" buffer)))))

(define (gl:frustum left right bottom top near-val far-val)
  (guarantee-current 'gl:draw-buffer)
  (guarantee flo:flonum? left 'gl:draw-buffer)
  (guarantee flo:flonum? right 'gl:draw-buffer)
  (guarantee flo:flonum? bottom 'gl:draw-buffer)
  (guarantee flo:flonum? top 'gl:draw-buffer)
  (guarantee flo:flonum? near-val 'gl:draw-buffer)
  (guarantee flo:flonum? far-val 'gl:draw-buffer)
  (C-call "glFrustum" left right bottom top near-val far-val))

(define (gl:gen-lists range)
  (guarantee-current 'gl:gen-lists)
  (guarantee integer? range 'gl:gen-lists)
  (C-call "glGenLists" range))

(define (gl:light light param values)
  (guarantee-current 'gl:light)
  (C-call "gl_light"
	  (case light
	    ((light0) (C-enum "GL_LIGHT0"))
	    ((light1) (C-enum "GL_LIGHT1"))
	    ((light2) (C-enum "GL_LIGHT2"))
	    ((light3) (C-enum "GL_LIGHT3"))
	    ((light4) (C-enum "GL_LIGHT4"))
	    ((light5) (C-enum "GL_LIGHT5"))
	    ((light6) (C-enum "GL_LIGHT6"))
	    ((light7) (C-enum "GL_LIGHT7"))
	    (else (error "gl:light: Unknown light:" light)))
	  (case param
	    ((position)
	     (guarantee-4d values 'gl:light)
	     (C-enum "GL_POSITION"))
	    ((ambient)
	     (guarantee-4d values 'gl:light)
	     (C-enum "GL_AMBIENT"))
	    ((diffuse)
	     (guarantee-4d values 'gl:light)
	     (C-enum "GL_DIFFUSE"))
	    ((specular)
	     (guarantee-4d values 'gl:light)
	     (C-enum "GL_SPECULAR"))
	    (else (error "gl:light: Unknown parameter:" param)))
	  values))

(define (gl:light-model param value)
  (case param
    ((local-viewer)
     (guarantee flo:flonum? value 'gl:light-model)
     (C-call "glLightModelf" (C-enum "GL_LIGHT_MODEL_LOCAL_VIEWER") value))
    ((color-control)
     (C-call "glLightModeli" (C-enum "GL_LIGHT_MODEL_COLOR_CONTROL")
	     (case value
	       ((separate-specular-color) (C-enum "GL_SEPARATE_SPECULAR_COLOR"))
	       ((single-color) (C-enum "GL_SINGLE_COLOR"))
	       (else (error "gl:light-model: Unknown color-control:" value)))))
    ((two-side)
     (guarantee flo:flonum? value 'gl:light-model)
     (C-call "glLightModelf" (C-enum "GL_LIGHT_MODEL_TWO_SIDE") value))
    ((ambient)
     (guarantee-4d value 'gl:light-model)
     (C-call "gl_light_model_v" (C-enum "GL_LIGHT_MODEL_AMBIENT") value))
    (else
     (error "gl:light-model: Unknown parameter:" param))))

(define (gl:material face param values)
  (guarantee-current 'gl:material)
  (C-call "gl_material"
	  (case face
	    ((front) (C-enum "GL_FRONT"))
	    ((back) (C-enum "GL_BACK"))
	    ((front-and-back) (C-enum "GL_FRONT_AND_BACK"))
	    (else (error "gl:material: Unknown face:" face)))
	  (case param
	    ((ambient)
	     (guarantee-4d values 'gl:material)
	     (C-enum "GL_AMBIENT"))
	    ((diffuse)
	     (guarantee-4d values 'gl:material)
	     (C-enum "GL_DIFFUSE"))
	    ((specular)
	     (guarantee-4d values 'gl:material)
	     (C-enum "GL_SPECULAR"))
	    ((emission)
	     (guarantee-4d values 'gl:material)
	     (C-enum "GL_EMISSION"))
	    ((shininess)
	     (guarantee flo:flonum? values 'gl:material)
	     (C-enum "GL_SHININESS"))
	    ((ambient-and-diffuse)
	     (guarantee-4d values 'gl:material)
	     (C-enum "GL_AMBIENT_AND_DIFFUSE"))
	    ((color-indexes)
	     (guarantee-3d values 'gl:material)
	     (C-enum "GL_COLOR_INDEXES"))
	    (else (error "gl:material: Unknown parameter:" param)))
	  values))

(define (gl:matrix-mode mode)
  (guarantee-current 'gl:matrix-mode)
  (C-call "glMatrixMode"
	  (case mode
	    ((modelview) (C-enum "GL_MODELVIEW"))
	    ((projection) (C-enum "GL_PROJECTION"))
	    (else (error "gl:matrix-mode: Unknown mode:" mode)))))

(define (gl:new-list lst mode)
  (guarantee-current 'gl:new-list)
  (guarantee integer? lst 'gl:new-list)
  (C-call "glNewList"
	  lst
	  (case mode
	    ((compile) (C-enum "GL_COMPILE"))
	    (else (error "gl:new-list: Unknown mode:" mode)))))

(define (gl:end-list)
  (C-call "glEndList"))

(define (gl:delete-lists lst range)
  (guarantee-current 'gl:delete-lists)
  (guarantee integer? lst 'gl:delete-lists)
  (guarantee integer? range 'gl:delete-lists)
  (C-call "glDeleteLists" lst range))

(define (gl:normal 3d)
  (guarantee-current 'gl:normal)
  (guarantee-3d 3d 'gl:normal)
  (C-call "glNormal3dv" 3d))

(define (gl:pop-matrix)
  (guarantee-current 'gl:pop-matrix)
  (C-call "glPopMatrix"))

(define (gl:push-matrix)
  (guarantee-current 'gl:push-matrix)
  (C-call "glPushMatrix"))

(define (gl:rotate angle x y z)
  (guarantee-current 'gl:rotate)
  (guarantee flo:flonum? x 'gl:rotate)
  (guarantee flo:flonum? y 'gl:rotate)
  (guarantee flo:flonum? z 'gl:rotate)
  (C-call "glRotated" angle x y z))

(define (gl:translate x y z)
  (guarantee-current 'gl:translate)
  (guarantee flo:flonum? x 'gl:translate)
  (guarantee flo:flonum? y 'gl:translate)
  (guarantee flo:flonum? z 'gl:translate)
  (C-call "glTranslated" x y z))

(define (gl:viewport x y width height)
  (guarantee-current 'gl:viewport)
  (guarantee integer? x 'gl:viewport)
  (guarantee integer? y 'gl:viewport)
  (guarantee integer? width 'gl:viewport)
  (guarantee integer? height 'gl:viewport)
  (C-call "glViewport" x y width height))

(define (gl:flush)
  (guarantee-current 'gl:flush)
  (C-call "glFlush"))

(define (glu:look-at position aim up)
  (guarantee-current 'gl:look-at)
  (C-call "glu_look_at" position aim up))

(define (glu:perspective fovy aspect z-near z-far)
  (guarantee-current 'gl:perspective)
  (C-call "gluPerspective" fovy aspect z-near z-far))

(define param:gl-context-current?)

(define (init)
  (set! param:gl-context-current? (make-parameter #f)))

(define (initialize-package!)
  (init)
  (add-event-receiver! event:after-restore reset-gl))

(define (reset-gl)
  (init))

(define (with-gl-context thunk)
  (if (param:gl-context-current?)
      (error "The GL library already has a context."))
  (parameterize ((param:gl-context-current? #t))
    (thunk)))

(define (guarantee-current operator)
  (if (not (param:gl-context-current?))
      (error "The GL library has no context:" operator)))

(define (guarantee-color object operator)
  (if (not (and (flo:flonum? object)
		(fix:= 4 (flo:vector-length object))))
      (error:wrong-type-argument object "a GL color" operator)))

(define (guarantee-gl-depth object operator)
  (if (not (and (flo:flonum? object)
		(<= 0.0 object) (<= object 1.0)))
      (error:wrong-type-argument object "a GL depth" operator)))

(define (guarantee-3d object operator)
  (if (not (and (flo:flonum? object)
		(fix:= 3 (flo:vector-length object))))
      (error:wrong-type-argument object "three flonums" operator)))

(define (guarantee-4d object operator)
  (if (not (and (flo:flonum? object)
		(fix:= 4 (flo:vector-length object))))
      (error:wrong-type-argument object "four flonums" operator)))

(initialize-package!)